# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import asyncio
import datetime
import pathlib

import pytest

import slim_bindings

keys_folder = f"{pathlib.Path(__file__).parent.resolve()}/testdata"

test_audience = ["test.audience"]


def create_slim(
    name: slim_bindings.PyName,
    private_key,
    private_key_algorithm,
    public_key,
    public_key_algorithm,
    wrong_audience=None,
):
    private_key = slim_bindings.PyKey(
        algorithm=private_key_algorithm,
        format=slim_bindings.PyKeyFormat.Pem,
        key=slim_bindings.PyKeyData.File(path=private_key),
    )

    public_key = slim_bindings.PyKey(
        algorithm=public_key_algorithm,
        format=slim_bindings.PyKeyFormat.Pem,
        key=slim_bindings.PyKeyData.File(path=public_key),
    )

    provider = slim_bindings.PyIdentityProvider.Jwt(
        private_key=private_key,
        duration=datetime.timedelta(seconds=60),
        issuer="test-issuer",
        audience=test_audience,
        subject=f"{name}",
    )
    verifier = slim_bindings.PyIdentityVerifier.Jwt(
        public_key=public_key,
        issuer="test-issuer",
        audience=wrong_audience or test_audience,
        require_iss=True,
        require_aud=True,
    )

    return slim_bindings.Slim.new(name, provider, verifier)


@pytest.mark.asyncio
@pytest.mark.parametrize("server", ["127.0.0.1:52345"], indirect=True)
@pytest.mark.parametrize("audience", [test_audience, ["wrong.audience"]])
async def test_identity_verification(server, audience):
    sender_name = slim_bindings.PyName("org", "default", "sender")
    receiver_name = slim_bindings.PyName("org", "default", "receiver")

    # Keys used for signing JWTs of sender
    private_key_sender = f"{keys_folder}/ec256.pem"
    public_key_sender = f"{keys_folder}/ec256-public.pem"
    algorithm_sender = slim_bindings.PyAlgorithm.ES256

    # Keys used for signing JWTs of receiver
    private_key_receiver = f"{keys_folder}/ec384.pem"
    public_key_receiver = f"{keys_folder}/ec384-public.pem"
    algorithm_receiver = slim_bindings.PyAlgorithm.ES384

    # create new slim object. note that the verifier will use the public key of the receiver
    # to verify the JWT of the reply message
    slim_sender = await create_slim(
        sender_name,
        private_key_sender,
        algorithm_sender,
        public_key_receiver,
        algorithm_receiver,
    )

    # Connect to the service and subscribe for the local name
    _ = await slim_sender.connect(
        {"endpoint": "http://127.0.0.1:52345", "tls": {"insecure": True}}
    )

    # create second local app. note that the receiver will use the public key of the sender
    # to verify the JWT of the request message
    slim_receiver = await create_slim(
        receiver_name,
        private_key_receiver,
        algorithm_receiver,
        public_key_sender,
        algorithm_sender,
        audience,
    )

    # Connect to SLIM server
    _ = await slim_receiver.connect(
        {"endpoint": "http://127.0.0.1:52345", "tls": {"insecure": True}}
    )

    # set route
    await slim_sender.set_route(receiver_name)

    # create request/reply session with default config
    session_info = await slim_sender.create_session(
        slim_bindings.PySessionConfiguration.FireAndForget(
            timeout=datetime.timedelta(seconds=1), max_retries=3, sticky=False
        )
    )

    # messages
    pub_msg = str.encode("thisistherequest")
    res_msg = str.encode("thisistheresponse")

    # Test with reply
    async with slim_sender, slim_receiver:
        # create background task for slim_receiver
        async def background_task():
            try:
                # wait for message from any new session
                recv_session, _ = await slim_receiver.receive()

                # receive message from session
                recv_session, msg_rcv = await slim_receiver.receive(
                    session=recv_session.id
                )

                # make sure the message is correct
                assert msg_rcv == bytes(pub_msg)

                # reply to the session
                await slim_receiver.publish_to(recv_session, res_msg)
            except Exception as e:
                print("Error receiving message on slim1:", e)

        t = asyncio.create_task(background_task())

        # send a request and expect a response in slim2
        if audience == test_audience:
            # As audience matches, we expect a successful request/reply
            session_info, message = await slim_sender.request_reply(
                session_info, pub_msg, receiver_name
            )

            # check if the message is correct
            assert message == bytes(res_msg)

            # Wait for task to finish
            await t
        else:
            # expect an exception due to audience mismatch
            with pytest.raises(asyncio.TimeoutError):
                session_info, message = await slim_sender.request_reply(
                    session_info,
                    pub_msg,
                    receiver_name,
                    timeout=datetime.timedelta(seconds=3),
                )

            # cancel the background task
            t.cancel()

            # wait for task to finish
            try:
                await t
            except asyncio.CancelledError:
                print("Background task cancelled as expected.")



# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
import typing
from enum import Enum, auto

class PyKey:
    algorithm: PyAlgorithm
    format: PyKeyFormat
    key: PyKeyData
    def __new__(cls,algorithm:PyAlgorithm, format:PyKeyFormat, key:PyKeyData): ...

class PyName:
    r"""
    name class
    """
    id: builtins.int
    def __new__(cls,component0:builtins.str, component1:builtins.str, component2:builtins.str, id:typing.Optional[builtins.int]=None): ...
    def set_id(self, id:builtins.int) -> None:
        ...

    def components(self) -> builtins.list[builtins.int]:
        ...

    def components_strings(self) -> builtins.list[builtins.str]:
        ...

    def equal_without_id(self, name:PyName) -> builtins.bool:
        ...

    def __repr__(self) -> builtins.str:
        ...

    def __str__(self) -> builtins.str:
        ...


class PyService:
    id: builtins.int
    name: PyName

class PySessionInfo:
    id: builtins.int
    source_name: PyName
    destination_name: PyName
    payload_type: builtins.str
    metadata: builtins.dict[builtins.str, builtins.str]
    def __new__(cls,session_id:builtins.int): ...

class PyAlgorithm(Enum):
    HS256 = auto()
    HS384 = auto()
    HS512 = auto()
    RS256 = auto()
    RS384 = auto()
    RS512 = auto()
    PS256 = auto()
    PS384 = auto()
    PS512 = auto()
    ES256 = auto()
    ES384 = auto()
    EdDSA = auto()

class PyIdentityProvider(Enum):
    StaticJwt = auto()
    Jwt = auto()
    SharedSecret = auto()

class PyIdentityVerifier(Enum):
    Jwt = auto()
    SharedSecret = auto()

class PyKeyData(Enum):
    File = auto()
    Content = auto()

class PyKeyFormat(Enum):
    Pem = auto()
    Jwk = auto()
    Jwks = auto()

class PySessionConfiguration(Enum):
    FireAndForget = auto()
    Streaming = auto()

class PySessionDirection(Enum):
    r"""
    session direction
    """
    SENDER = auto()
    RECEIVER = auto()
    BIDIRECTIONAL = auto()

class PySessionType(Enum):
    r"""
    session type
    """
    FIRE_AND_FORGET = auto()
    STREAMING = auto()

def connect(svc:PyService, config:dict) -> typing.Any:
    ...

def create_pyservice(name:PyName, provider:PyIdentityProvider, verifier:PyIdentityVerifier) -> typing.Any:
    ...

def create_session(svc:PyService, config:PySessionConfiguration) -> typing.Any:
    ...

def delete_session(svc:PyService, session_id:builtins.int) -> typing.Any:
    ...

def disconnect(svc:PyService, conn:builtins.int) -> typing.Any:
    ...

def get_default_session_config(svc:PyService, session_type:PySessionType) -> typing.Any:
    ...

def get_session_config(svc:PyService, session_id:builtins.int) -> typing.Any:
    ...

def init_tracing(config:dict) -> typing.Any:
    ...

def invite(svc:PyService, session_info:PySessionInfo, name:PyName) -> typing.Any:
    ...

def publish(svc:PyService, session_info:PySessionInfo, fanout:builtins.int, blob:typing.Sequence[builtins.int], name:typing.Optional[PyName]=None, payload_type:typing.Optional[builtins.str]=None, metadata:typing.Optional[typing.Mapping[builtins.str, builtins.str]]=None) -> typing.Any:
    ...

def receive(svc:PyService) -> typing.Any:
    ...

def remove(svc:PyService, session_info:PySessionInfo, name:PyName) -> typing.Any:
    ...

def remove_route(svc:PyService, conn:builtins.int, name:PyName) -> typing.Any:
    ...

def run_server(svc:PyService, config:dict) -> typing.Any:
    ...

def set_default_session_config(svc:PyService, config:PySessionConfiguration) -> typing.Any:
    ...

def set_route(svc:PyService, conn:builtins.int, name:PyName) -> typing.Any:
    ...

def set_session_config(svc:PyService, session_id:builtins.int, config:PySessionConfiguration) -> typing.Any:
    ...

def stop_server(svc:PyService, endpoint:builtins.str) -> typing.Any:
    ...

def subscribe(svc:PyService, conn:builtins.int, name:PyName) -> typing.Any:
    ...

def unsubscribe(svc:PyService, conn:builtins.int, name:PyName) -> typing.Any:
    ...




# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import argparse
import asyncio
from signal import SIGINT

import slim_bindings

from .common import shared_secret_identity


async def run_server(address: str, enable_opentelemetry: bool):
    # init tracing
    await slim_bindings.init_tracing(
        {
            "log_level": "debug",
            "opentelemetry": {
                "enabled": enable_opentelemetry,
                "grpc": {
                    "endpoint": "http://localhost:4317",
                },
            },
        }
    )

    global slim
    # not used in the slim server
    provider, verifier = shared_secret_identity(
        identity="slim",
        secret="secret",
    )
    # create new slim object
    slim = await slim_bindings.Slim.new(
        slim_bindings.PyName("cisco", "default", "slim"), provider, verifier
    )

    # Run as server
    await slim.run_server({"endpoint": address, "tls": {"insecure": True}})


async def amain():
    parser = argparse.ArgumentParser(description="Command line client for slim server.")
    parser.add_argument(
        "-s", "--slim", type=str, help="Slim address.", default="127.0.0.1:12345"
    )
    parser.add_argument(
        "--enable-opentelemetry",
        "-t",
        action="store_true",
        default=False,
        help="Enable OpenTelemetry tracing.",
    )

    args = parser.parse_args()

    # Create an asyncio event to keep the loop running until interrupted
    stop_event = asyncio.Event()

    # Define a shutdown handler to set the event when interrupted
    def shutdown():
        print("\nShutting down...")
        stop_event.set()

    # Register the shutdown handler for SIGINT
    loop = asyncio.get_running_loop()
    loop.add_signal_handler(SIGINT, shutdown)

    # Run the client task
    client_task = asyncio.create_task(run_server(args.slim, args.enable_opentelemetry))

    # Wait until the stop event is set
    await stop_event.wait()

    # Cancel the client task
    client_task.cancel()
    try:
        await client_task
    except asyncio.CancelledError:
        pass


def main():
    try:
        asyncio.run(amain())
    except KeyboardInterrupt:
        print("Program terminated by user.")



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import base64
import json

import click

import slim_bindings


class color:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


# Format a message for display
def format_message(message1, message2=""):
    return f"{color.BOLD}{color.CYAN}{message1.capitalize():<45}{color.END}{message2}"


def format_message_print(message1, message2=""):
    print(format_message(message1, message2))


# Split an ID into its components
# Expected format: organization/namespace/application
# Raises ValueError if the format is incorrect
# Returns a PyName with the 3 components
def split_id(id):
    try:
        organization, namespace, app = id.split("/")
    except ValueError as e:
        print("Error: IDs must be in the format organization/namespace/app-or-stream.")
        raise e

    return slim_bindings.PyName(organization, namespace, app)


# Create a shared secret identity provider and verifier
# This is used for shared secret authentication
# Takes an identity and a shared secret as parameters
# Returns a tuple of (provider, verifier)
# This is used for shared secret authentication
def shared_secret_identity(identity, secret):
    """
    Create a provider and verifier using a shared secret.
    """
    provider = slim_bindings.PyIdentityProvider.SharedSecret(
        identity=identity, shared_secret=secret
    )
    verifier = slim_bindings.PyIdentityVerifier.SharedSecret(
        identity=identity, shared_secret=secret
    )

    return provider, verifier


# Create a JWT identity provider and verifier
# This is used for JWT authentication
# Takes private key path, public key path, and algorithm as parameters
# Returns a Slim object with the provider and verifier
def jwt_identity(
    jwt_path: str,
    jwk_path: str,
    iss: str = None,
    sub: str = None,
    aud: list = None,
):
    """
    Parse the JWK and JWT from the provided strings.
    """

    print(f"Using JWk file: {jwk_path}")

    with open(jwk_path) as jwk_file:
        jwk_string = jwk_file.read()

    # The JWK is normally encoded as base64, so we need to decode it
    spire_jwks = json.loads(jwk_string)

    for _, v in spire_jwks.items():
        # Decode first item from base64
        spire_jwks = base64.b64decode(v)
        break

    provider = slim_bindings.PyIdentityProvider.StaticJwt(
        path=jwt_path,
    )

    pykey = slim_bindings.PyKey(
        algorithm=slim_bindings.PyAlgorithm.RS256,
        format=slim_bindings.PyKeyFormat.Jwks,
        key=slim_bindings.PyKeyData.Content(content=spire_jwks.decode("utf-8")),
    )

    verifier = slim_bindings.PyIdentityVerifier.Jwt(
        public_key=pykey,
        issuer=iss,
        audience=[aud],
        subject=sub,
    )

    return provider, verifier


# A custom click parameter type for parsing dictionaries from JSON strings
# This is useful for passing complex configurations via command line arguments
class DictParamType(click.ParamType):
    name = "dict"

    def convert(self, value, param, ctx):
        import json

        if isinstance(value, dict):
            return value  # Already a dict (for default value)
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            self.fail(f"{value} is not valid JSON", param, ctx)


def common_options(function):
    function = click.command(context_settings={"auto_envvar_prefix": "SLIM"})(function)

    function = click.option(
        "--local",
        type=str,
        required=True,
        help="Local ID in the format organization/namespace/application",
    )(function)

    function = click.option(
        "--remote",
        type=str,
        help="Remote ID in the format organization/namespace/application-or-stream",
    )(function)

    function = click.option(
        "--slim",
        default={
            "endpoint": "http://127.0.0.1:46357",
            "tls": {
                "insecure": True,
            },
        },
        type=DictParamType(),
        help="slim connection parameters",
    )(function)

    function = click.option(
        "--enable-opentelemetry",
        is_flag=True,
        help="Enable OpenTelemetry tracing",
    )(function)

    function = click.option(
        "--shared-secret",
        type=str,
        help="Shared secret for authentication. Don't use this in production.",
    )(function)

    function = click.option(
        "--jwt",
        type=str,
        help="JWT token for authentication.",
    )(function)

    function = click.option(
        "--bundle",
        type=str,
        help="Key bundle for authentication, in JWKS format.",
    )(function)

    function = click.option(
        "--audience",
        type=str,
        help="Audience for the JWT.",
    )(function)

    function = click.option(
        "--invites",
        type=str,
        multiple=True,
        help="Invite other participants to the pubsub session. Can be specified multiple times.",
    )(function)

    function = click.option(
        "--enable-mls",
        is_flag=True,
        help="Enable MLS (Message Layer Security) for the pubsub session.",
    )(function)

    return function


async def create_local_app(
    local: str,
    slim: dict,
    remote: str | None = None,
    enable_opentelemetry: bool = False,
    shared_secret: str | None = None,
    jwt: str | None = None,
    bundle: str | None = None,
    audience: list[str] | None = None,
):
    # init tracing
    slim_bindings.init_tracing(
        {
            "log_level": "info",
            "opentelemetry": {
                "enabled": enable_opentelemetry,
                "grpc": {
                    "endpoint": "http://localhost:4317",
                },
            },
        }
    )

    if not jwt and not bundle:
        if not shared_secret:
            raise ValueError(
                "Either JWT or bundle must be provided, or a shared secret."
            )

    # Derive identity provider and verifier from JWK and JWT
    if jwt and bundle:
        provider, verifier = jwt_identity(
            jwt,
            bundle,
            aud=audience,
        )
    else:
        provider, verifier = shared_secret_identity(
            identity=local,
            secret=shared_secret,
        )

    # Split the local IDs into their respective components
    local_name = split_id(local)

    local_app = await slim_bindings.Slim.new(local_name, provider, verifier)

    format_message_print(f"{local_app.get_id()}", "Created app")

    # Connect to slim server
    _ = await local_app.connect(slim)

    format_message_print(f"{local_app.get_id()}", f"Connected to {slim['endpoint']}")

    return local_app



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import slim_bindings


async def create_svc(name: slim_bindings.PyName, secret):
    provider = slim_bindings.PyIdentityProvider.SharedSecret(
        identity=f"{name}", shared_secret=secret
    )
    verifier = slim_bindings.PyIdentityVerifier.SharedSecret(
        identity=f"{name}", shared_secret=secret
    )
    return await slim_bindings.create_pyservice(name, provider, verifier)


async def create_slim(name: slim_bindings.PyName, secret):
    provider = slim_bindings.PyIdentityProvider.SharedSecret(
        identity=f"{name}", shared_secret=secret
    )
    verifier = slim_bindings.PyIdentityVerifier.SharedSecret(
        identity=f"{name}", shared_secret=secret
    )
    return await slim_bindings.Slim.new(name, provider, verifier)



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import asyncio
import datetime

import slim_bindings

from .common import (
    common_options,
    create_local_app,
    format_message_print,
    split_id,
)


async def run_client(
    local: str,
    slim: dict,
    remote: str | None,
    enable_opentelemetry: bool = False,
    enable_mls: bool = False,
    shared_secret: str | None = None,
    jwt: str | None = None,
    bundle: str | None = None,
    audience: list[str] | None = None,
    invites: list[str] | None = None,
):
    local_app = await create_local_app(
        local,
        slim,
        enable_opentelemetry=enable_opentelemetry,
        shared_secret=shared_secret,
        jwt=jwt,
        bundle=bundle,
        audience=audience,
    )

    # If provided, split the remote IDs into their respective components
    if remote:
        broadcast_topic = split_id(remote)

    tasks = []

    if remote and invites:
        format_message_print(local, "Creating new pubsub sessions...")
        # create pubsubb session. A pubsub session is a just a bidirectional
        # streaming session, where participants are both sender and receivers
        session_info = await local_app.create_session(
            slim_bindings.PySessionConfiguration.Streaming(
                slim_bindings.PySessionDirection.BIDIRECTIONAL,
                topic=broadcast_topic,
                moderator=True,
                max_retries=5,
                timeout=datetime.timedelta(seconds=5),
                mls_enabled=enable_mls,
            )
        )

        # invite all participants
        for p in invites:
            to_add = split_id(p)
            await local_app.set_route(to_add)
            await local_app.invite(session_info, to_add)
            print(f"{local} -> add {to_add} to the group")

    # define the background task
    async def background_task():
        async with local_app:
            # init session from session
            if invites:
                recv_session = session_info
            else:
                format_message_print(local, "-> Waiting for session...")
                recv_session, _ = await local_app.receive()

            while True:
                try:
                    # receive message from session
                    recv_session, msg_rcv = await local_app.receive(
                        session=recv_session.id
                    )

                    # print received message
                    format_message_print(
                        local,
                        f"-> Received message from {recv_session.id}: {msg_rcv.decode()}",
                    )

                    # Here we could send a message back to the channel, e.g.:
                    # await participant.publish(session_info, f"Echo: {msg_rcv.decode()}".encode())
                except asyncio.CancelledError:
                    break
                except Exception as e:
                    format_message_print(local, f"-> Error receiving message: {e}")
                    break

    tasks.append(asyncio.create_task(background_task()))

    if remote and invites:

        async def background_task_keyboard():
            while True:
                user_input = await asyncio.to_thread(input, "\033[1mmessage>\033[0m ")
                if user_input == "exit":
                    break

                # Send the message to the all participants
                await local_app.publish(
                    session_info,
                    f"{user_input}".encode(),
                    broadcast_topic,
                )

        tasks.append(asyncio.create_task(background_task_keyboard()))

    # Wait for both tasks to finish
    await asyncio.gather(*tasks)


@common_options
def main(
    local: str,
    slim: dict,
    remote: str | None = None,
    enable_opentelemetry: bool = False,
    enable_mls: bool = False,
    shared_secret: str | None = None,
    jwt: str | None = None,
    bundle: str | None = None,
    audience: list[str] | None = None,
    invites: list[str] | None = None,
):
    try:
        asyncio.run(
            run_client(
                local=local,
                slim=slim,
                remote=remote,
                enable_opentelemetry=enable_opentelemetry,
                enable_mls=enable_mls,
                shared_secret=shared_secret,
                jwt=jwt,
                bundle=bundle,
                audience=audience,
                invites=invites,
            )
        )
    except KeyboardInterrupt:
        print("Client interrupted by user.")



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import asyncio
import datetime

import pytest
from common import create_slim

import slim_bindings


@pytest.mark.asyncio
@pytest.mark.parametrize("server", ["127.0.0.1:12356"], indirect=True)
async def test_request_reply(server):
    org = "cisco"
    ns = "default"
    app1 = "slim1"

    name1 = slim_bindings.PyName(org, ns, app1)

    # create new slim object
    slim1 = await create_slim(name1, "secret")

    # Connect to the service and subscribe for the local name
    _ = await slim1.connect(
        {"endpoint": "http://127.0.0.1:12356", "tls": {"insecure": True}}
    )

    # create second local app
    name2 = slim_bindings.PyName(org, ns, "slim2")
    slim2 = await create_slim(name2, "secret")

    # Connect to SLIM server
    _ = await slim2.connect(
        {"endpoint": "http://127.0.0.1:12356", "tls": {"insecure": True}}
    )

    # set route
    await slim2.set_route(name1)

    # create request/reply session with default config
    session_info = await slim2.create_session(
        slim_bindings.PySessionConfiguration.FireAndForget(
            timeout=datetime.timedelta(seconds=1), max_retries=3, sticky=False
        )
    )

    # messages
    pub_msg = str.encode("thisistherequest")
    res_msg = str.encode("thisistheresponse")

    # Test with reply
    async with slim1, slim2:
        # create background task for slim1
        async def background_task():
            try:
                # wait for message from any new session
                recv_session, _ = await slim1.receive()

                # receive message from session
                recv_session, msg_rcv = await slim1.receive(session=recv_session.id)

                # make sure the message is correct
                assert msg_rcv == bytes(pub_msg)
                assert recv_session.destination_name.equal_without_id(name1)

                # reply to the session
                await slim1.publish_to(recv_session, res_msg)
            except Exception as e:
                print("Error receiving message on slim1:", e)

        t = asyncio.create_task(background_task())

        # send a request and expect a response in slim2
        session_info, message = await slim2.request_reply(session_info, pub_msg, name1)

        # check if the message is correct
        assert message == bytes(res_msg)
        assert session_info.destination_name.equal_without_id(name2)

        # wait for task to finish
        await t

    # Test without reply
    async with slim1, slim2:
        # send a request. No one is listening, so expect a timeout exception
        with pytest.raises(asyncio.TimeoutError):
            session_info, message = await slim2.request_reply(
                session_info,
                pub_msg,
                name1,
                timeout=datetime.timedelta(seconds=5),
            )



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import asyncio
import datetime

import pytest
from common import create_slim

import slim_bindings


@pytest.mark.asyncio
@pytest.mark.parametrize("server", ["127.0.0.1:22345"], indirect=True)
@pytest.mark.parametrize("mls_enabled", [True, False])
async def test_sticky_session(server, mls_enabled):
    sender_name = slim_bindings.PyName("org", "default", "sender")
    receiver_name = slim_bindings.PyName("org", "default", "receiver")

    # create new slim object
    sender = await create_slim(sender_name, "secret")

    # Connect to the service and subscribe for the local name
    _ = await sender.connect(
        {"endpoint": "http://127.0.0.1:22345", "tls": {"insecure": True}}
    )

    # set route to receiver
    await sender.set_route(receiver_name)

    receiver_counts = {i: 0 for i in range(10)}

    # run 10 receivers concurrently
    async def run_receiver(i: int):
        # create new receiver object
        receiver = await create_slim(receiver_name, "secret")

        # Connect to the service and subscribe for the local name
        _ = await receiver.connect(
            {"endpoint": "http://127.0.1:22345", "tls": {"insecure": True}}
        )

        async with receiver:
            # wait for a new session
            session_info_rec, _ = await receiver.receive()

            print(f"Receiver {i} received session: {session_info_rec.id}")

            # new session received! listen for the message
            while True:
                info, _ = await receiver.receive(session=session_info_rec.id)

                if (
                    info.destination_name.equal_without_id(receiver_name)
                    and info.payload_type == "hello message"
                    and info.metadata.get("sender") == "hello"
                ):
                    # store the count in dictionary
                    receiver_counts[i] += 1

    # run 10 receivers concurrently
    tasks = []
    for i in range(10):
        t = asyncio.create_task(run_receiver(i))
        tasks.append(t)
        await asyncio.sleep(0.1)

    # create a new session
    session_info = await sender.create_session(
        slim_bindings.PySessionConfiguration.FireAndForget(
            max_retries=5,
            timeout=datetime.timedelta(seconds=5),
            sticky=True,
            mls_enabled=mls_enabled,
        )
    )

    # Wait a moment
    await asyncio.sleep(2)

    payload_type = "hello message"
    metadata = {"sender": "hello"}

    # send a message to the receiver
    for i in range(1000):
        await sender.publish(
            session_info, b"Hello from sender", receiver_name, payload_type, metadata
        )

    # Wait for all receivers to finish
    await asyncio.sleep(1)

    # As we setup a sticky session, all the message should be received by only one
    # receiver. Check that the count is 1000 for one of the receivers
    assert 1000 in receiver_counts.values()

    # Kill all tasks
    for task in tasks:
        task.cancel()



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import asyncio

import pytest_asyncio

import slim_bindings


@pytest_asyncio.fixture(scope="function")
async def server(request):
    # create new server
    global svc_server

    name = slim_bindings.PyName("agntcy", "default", "server")
    provider = slim_bindings.PyIdentityProvider.SharedSecret(
        identity="server", shared_secret="secret"
    )
    verifier = slim_bindings.PyIdentityVerifier.SharedSecret(
        identity="server", shared_secret="secret"
    )

    svc_server = await slim_bindings.create_pyservice(name, provider, verifier)

    # init tracing
    await slim_bindings.init_tracing({"log_level": "info"})

    # run slim server in background
    await slim_bindings.run_server(
        svc_server,
        {"endpoint": request.param, "tls": {"insecure": True}},
    )

    # wait for the server to start
    await asyncio.sleep(1)

    # return the server
    yield svc_server



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

from .fire_and_forget import main as fire_and_forget_main
from .pubsub import main as pubsub_main
from .request_reply import main as request_reply_main
from .slim import main as slim_main

HELP = """
This is the slim bindings examples package.
Available commands:
    - fire-and-forget: Demonstrates fire-and-forget messaging.
    - request-reply: Demonstrates request-reply messaging.
    - pubsub: Demonstrates publish-subscribe messaging.
    - slim: Starts a SLIM instance.

Use 'slim-bindings-examples <command>' to run a specific example.
For example: 'slim-bindings-examples fire-and-forget'.
"""


def main():
    # Check what command was provided
    import sys

    if len(sys.argv) > 1:
        command = sys.argv[1]
        sys.argv = sys.argv[1:]

        if command == "ff":
            fire_and_forget_main()
        elif command == "rr":
            request_reply_main()
        elif command == "pubsub":
            pubsub_main()
        elif command == "slim":
            slim_main()
        else:
            print(f"Unknown command: {command}")
            print(HELP)
    else:
        print("No command provided.")
        print(HELP)


if __name__ == "__main__":
    main()



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import asyncio
import datetime

import click

import slim_bindings

from .common import (
    common_options,
    create_local_app,
    format_message_print,
    split_id,
)


async def run_client(
    local: str,
    slim: dict,
    remote: str | None,
    enable_opentelemetry: bool = False,
    enable_mls: bool = False,
    shared_secret: str | None = None,
    jwt: str | None = None,
    bundle: str | None = None,
    audience: list[str] | None = None,
    message: str | None = None,
    iterations: int = 1,
    sticky: bool = False,
):
    local_app: slim_bindings.Slim = await create_local_app(
        local,
        slim,
        enable_opentelemetry=enable_opentelemetry,
        shared_secret=shared_secret,
        jwt=jwt,
        bundle=bundle,
        audience=audience,
    )

    instance = local_app.get_id()
    local_name = split_id(local)

    if message:
        if not remote:
            raise ValueError("Remote ID must be provided when message is specified.")

    async with local_app:
        if message:
            # Split the IDs into their respective components
            remote_name = split_id(remote)

            # Create a route to the remote ID
            await local_app.set_route(remote_name)

            # create a session
            if sticky or enable_mls:
                session = await local_app.create_session(
                    slim_bindings.PySessionConfiguration.FireAndForget(
                        max_retries=5,
                        timeout=datetime.timedelta(seconds=5),
                        sticky=True,
                        mls_enabled=enable_mls,
                    )
                )
            else:
                session = await local_app.create_session(
                    slim_bindings.PySessionConfiguration.FireAndForget()
                )

            for i in range(0, iterations):
                try:
                    # Send the message
                    await local_app.publish(
                        session,
                        message.encode(),
                        remote_name,
                    )

                    format_message_print(
                        f"{instance}",
                        f"Sent message {message} - {i + 1}/{iterations}:",
                    )

                    # Wait for a reply
                    session_info, msg = await local_app.receive(session=session.id)

                    format_message_print(
                        f"{instance}",
                        f"received (from session {session_info.id}): {msg.decode()}",
                    )

                    if not session_info.destination_name.equal_without_id(local_name):
                        format_message_print(
                            f"received message with wrong name, exit. local {local_name}, dst {session_info.destination_name}"
                        )
                        exit(1)

                except Exception as e:
                    print("received error: ", e)

                await asyncio.sleep(1)
        else:
            # Wait for a message and reply in a loop
            while True:
                format_message_print(
                    f"{instance}",
                    "waiting for new session to be established",
                )

                session_info, _ = await local_app.receive()
                format_message_print(
                    f"{instance} received a new session:",
                    f"{session_info.id}",
                )

                async def background_task(session_id):
                    while True:
                        # Receive the message from the session
                        session, msg = await local_app.receive(session=session_id)
                        format_message_print(
                            f"{instance}",
                            f"received (from session {session_id}): {msg.decode()}",
                        )

                        if not session.destination_name.equal_without_id(local_name):
                            format_message_print(
                                f"received message with wrong name, exit. local {local_name}, dst {session.destination_name}"
                            )
                            exit(1)

                        ret = f"{msg.decode()} from {instance}"

                        await local_app.publish_to(session, ret.encode())
                        format_message_print(f"{instance}", f"replies: {ret}")

                asyncio.create_task(background_task(session_info.id))


def ff_options(function):
    function = click.option(
        "--message",
        type=str,
        help="Message to send.",
    )(function)

    function = click.option(
        "--iterations",
        type=int,
        help="Number of messages to send, one per second.",
        default=2,
    )(function)

    function = click.option(
        "--sticky",
        is_flag=True,
        help="Enable FF sessions to connect always to the same endpoint.",
        default=False,
    )(function)

    return function


@common_options
@ff_options
def main(
    local: str,
    slim: dict,
    remote: str | None = None,
    enable_opentelemetry: bool = False,
    enable_mls: bool = False,
    shared_secret: str | None = None,
    jwt: str | None = None,
    bundle: str | None = None,
    audience: list[str] | None = None,
    invites: list[str] | None = None,
    message: str | None = None,
    iterations: int = 1,
    sticky: bool = False,
):
    try:
        asyncio.run(
            run_client(
                local=local,
                slim=slim,
                remote=remote,
                enable_opentelemetry=enable_opentelemetry,
                enable_mls=enable_mls,
                shared_secret=shared_secret,
                jwt=jwt,
                bundle=bundle,
                audience=audience,
                message=message,
                iterations=iterations,
                sticky=sticky,
            )
        )
    except KeyboardInterrupt:
        print("Client interrupted by user.")
